package de.unibielefeld.cebitec.gi.algogr.exl2;

import de.unibielefeld.cebitec.gi.algogr.exl2.misc.HiddenMarkovSampler;
import de.unibielefeld.cebitec.gi.algogr.exl2.misc.Utils;

/**
 * Exercise List 2.
 */
public class ExerciseList2 {

    private Viterbi<Character, Integer> viterbi;
    private Forward<Character, Integer> forwardAlg;
    private Backward<Character, Integer> backwardAlg;

    private HiddenMarkovSampler<Character, Integer> sampler;
    private HMM<Character, Integer> casinoHMM;

    public void doExercise() {
        //
        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // BEGIN OF EXERCISE LIST 2
        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        //

        /*
         * Constructing HMM of the occasionally dishonest casino.
         */

        // alphabet
        Integer[] sigma = new Integer[] { 1, 2, 3, 4, 5, 6 };

        // set of states
        Character[] Q = new Character[] { 'F', 'U' };

        // state transition matrix
        double[][] A = new double[][] {
            // transition probabilities for 'Fair' state
            {
                // F -> F
                0.95,
                // F -> U
                0.05 },
            // transition probabilities for 'Unfair' state
            {
                // U -> F
                0.1,
                // U -> U
                0.9 } };

        // matrix of emission probabilities
        double[][] E = new double[][] {
            // fair die
            //
            // 1 2 3 4 5 6
            { 1d / 6, 1d / 6, 1d / 6, 1d / 6, 1d / 6, 1d / 6 },

            // unfair die
            //
            // 1 2 3 4 5 6
            { 1d / 10, 1d / 10, 1d / 10, 1d / 10, 1d / 10, 1d / 2 } };

        // transition probabilities from begin state
        double[] alpha = new double[] {
            // begin -> F
            0.5,
            // begin -> U
            0.5 };

        // instantiate HMM of the occasionally dishonest casino.
        this.casinoHMM = new HMM<Character, Integer>(sigma, Q, A, E, alpha);

        /*
         * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
         * Exercise 3 (a)
         * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
         */

        // emission sequence x
        Integer[] x = new Integer[] { 5, 1, 6, 5, 4, 5, 5, 4, 2 };

        // TODO compute P(x)

        /*
         * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
         * Exercise 3 (b)
         * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
         */

        // instantiate sequence generator for HMM of the occasionally dishonest
        // casino.
        this.sampler = new HiddenMarkovSampler<>(this.casinoHMM);

        // generate state path and emission sequence of a specified length

        // TODO call sample() function of the HiddenMarkovSampler
        HiddenMarkovSampler.Sample<Character, Integer> sample = null;

        // TODO obtain state path and emission sequence from sample
        Character[] truePath = new Character[] {}; // dummy;
        Integer[] sequence = new Integer[] {}; // dummy;

        // TODO run the Viterbi Algorithm on the generated emission
        // sequence

        // TODO: the class Viterbi already implements the method
        // "getMostProbablePath", which, once the ptr table (see "run" method of
        // Viterbi) is correctly filled, backtraces the most probable path!
        Character[] mostProbPath = null;

        // TODO print the estimated sequence, true state path and most probable
        // state path returned by Viterbi Algorithm by uncommenting the
        // following line:

        // Utils.printEstimatePath(sequence, truePath, mostProbPath);

        /*
         * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
         * Exercise 3 (c)
         * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
         */

        // TODO run the forward and backward algorithm on the emission sequence
        // generated in Exercise 3 (b)

        // TODO compute the posterior distribution using the variables of the
        // forward and backward algorithm and P(x)

        double[][] posteriorDistribution = new double[sequence.length][Q.length];

        // plot the posterior distribution for the state corresponding to the
        // fair die
        Utils.plotPosteriorDistributionForStateAndPath(posteriorDistribution,
            this.casinoHMM.indexOfState('F'), truePath,
            "Posterior Distribution of the Fair Die");
    }

    //
    // END OF EXERCISE LIST 2
    //

    public ExerciseList2() {
        this.viterbi = new Viterbi<Character, Integer>();
        this.forwardAlg = new Forward<Character, Integer>();
        this.backwardAlg = new Backward<Character, Integer>();
    }

    public static void main(String[] args) {
        ExerciseList2 exl2 = new ExerciseList2();
        exl2.doExercise();
    }
}
