package de.unibielefeld.cebitec.gi.algogr.exl2;

import java.util.Arrays;

/**
 * Viterbi Algorithm.
 * 
 * @param <State>
 *            State Type
 * @param <Alphabet>
 *            Emission alphabet type
 */
public class Viterbi<State, Alphabet> {

    private double[][] v;
    private int[][] ptr;

    private HMM<State, Alphabet> hmm;

    /**
     * Method that implements the Viterbi Algorithm.
     * 
     * @param hmm
     *            HMM
     * @param x
     *            emission sequence
     * @return Probability P(x, pi*) of emitting x by state path pi*
     */
    public double run(HMM<State, Alphabet> hmm, Alphabet[] x) {

        this.hmm = hmm;

        // access required HMM data structures
        double[][] A = hmm.getTransitionMatrix();
        double[][] E = hmm.getEmissionMatrix();
        State[] Q = hmm.getStates();
        double[] alpha = hmm.getBeginDistribution();

        /*
         * VITERBI ALGORITHM
         */

        // initialize table of v variables
        this.v = new double[x.length][Q.length];
        // initialize pointer table for backtracing the most probable path
        this.ptr = new int[x.length][Q.length];

        /*
         * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
         * Exercise 2: Implement the Viterbi algorithm by correctly filling the
         * initialized table of v variables and the pointer table ptr!
         * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
         */

        // START

        // INITIALIZATION (i=0)
        // TODO implement the following statements:
        // v_{begin state} (1) = 1, v_k(0) = 0 for k in Q
        //
        // ptr_k(1) = -1 for k in Q (this initializes the ptr table for the
        // first position to a default value)

        // RECURSION (i=1,...,L)
        // TODO implement the following statements:
        // v_l(i) = e_l(x_i) * max_k(v_k(i-1) * a_{k,l})
        //
        // ptr_i(l) = argmax_k(v_k(i-1) * a_{k,l})

        // TERMINATION
        // TODO implement the following statement:
        // P(x, pi*) = max_k(v_k(L))
        double P_x_pi_star = 0;

        // END

        // return P(x, pi*)
        return P_x_pi_star;
    }

    /**
     * Method to obtain the most probable path by backtracing the ptr table.
     * 
     * @return most probable state path corresponding to emission sequence x
     */
    public State[] getMostProbablePath() {

        State[] Q = this.hmm.getStates();
        State[] res = Arrays.copyOf(Q, this.v.length);

        int prev_k = 0;

        for (int k = 1; k < this.hmm.getStates().length; k++) {
            if (this.v[this.v.length - 1][k] > this.v[this.v.length - 1][prev_k]) {
                prev_k = k;
            }
        }

        res[res.length - 1] = Q[prev_k];

        for (int i = this.ptr.length - 1; i > 0; i--) {
            res[i - 1] = Q[this.ptr[i][prev_k]];
            prev_k = this.ptr[i][prev_k];
        }
        return res;
    }

    /**
     * Getter for v variables.
     * 
     * @return v variables computed in run() function
     */
    public double[][] getVariables() {
        return this.v;
    }
}
